from __future__ import annotations

import asyncio
import json
import logging
import pathlib
import re
import shlex
from typing import Any

import yaml

from dagster_uc._helm.errors import (
    ChartNotFoundError,
    CommandCancelledError,
    Error,
    FailedToRenderChartError,
    HelmConnectionError,
    InvalidResourceError,
    ReleaseNotFoundError,
    ResourceAlreadyExistsError,
)


class SafeLoader(yaml.SafeLoader):
    """We use a custom YAML loader that doesn't bork on plain equals '=' signs.

    It was originally designated with a special meaning, but noone uses it:

        https://github.com/yaml/pyyaml/issues/89
        https://yaml.org/type/value.html
    """

    @staticmethod
    def construct_value(loader, node):  # noqa
        return loader.construct_scalar(node)


SafeLoader.add_constructor("tag:yaml.org,2002:value", SafeLoader.construct_value)


CHART_NOT_FOUND = re.compile(r"chart \"[^\"]+\" (version \"[^\"]+\" )?not found")
CONNECTION_ERROR = re.compile(r"(read: operation timed out|connect: network is unreachable)")


class Command:
    """Class presenting an async interface around the Helm CLI."""

    def __init__(
        self,
        *,
        default_timeout: int | str = "5m",
        executable: str = "helm",
        history_max_revisions: int = 10,
        insecure_skip_tls_verify: bool = False,
        kubeconfig: pathlib.Path | None = None,
        kubecontext: str | None = None,
        unpack_directory: str | None = None,
    ):
        self._logger = logging.getLogger(__name__)
        self._default_timeout = default_timeout
        self._executable = executable
        self._history_max_revisions = history_max_revisions
        self._insecure_skip_tls_verify = insecure_skip_tls_verify
        self._kubeconfig = kubeconfig
        self._kubecontext = kubecontext
        self._unpack_directory = unpack_directory

    def _log_format(self, argument):  # noqa: ANN001, ANN202
        argument = str(argument)
        if argument == "-":
            return "<stdin>"
        elif "\n" in argument:
            return "<multi-line string>"
        else:
            return argument

    async def run(self, command: list[str], input: bytes | None = None) -> bytes:  # noqa: A002
        """Run the given Helm command with the given input as stdin and"""
        command = [self._executable] + command
        if self._kubeconfig:
            command.extend(["--kubeconfig", self._kubeconfig.absolute().as_posix()])
        if self._kubecontext:
            command.extend(["--kube-context", self._kubecontext])
        # The command must be made up of str and bytes, so convert anything that isn't
        shell_formatted_command = shlex.join(
            part if isinstance(part, str | bytes) else str(part) for part in command
        )
        log_formatted_command = shlex.join(self._log_format(part) for part in command)
        self._logger.info("running command: %s", log_formatted_command)
        proc = await asyncio.create_subprocess_shell(
            shell_formatted_command,
            # Only make stdin a pipe if we have input to feed it
            stdin=asyncio.subprocess.PIPE if input is not None else None,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        try:
            stdout, stderr = await proc.communicate(input)
        except asyncio.CancelledError:
            # If the asyncio task is cancelled, terminate the Helm process but let the
            # process handle the termination and exit
            # We occasionally see a ProcessLookupError here if the process finished between
            # us being cancelled and terminating the process, which we ignore as that is our
            # target state anyway
            try:
                proc.terminate()
                _ = await proc.communicate()
            except ProcessLookupError:
                pass
            # Once the process has exited, re-raise the cancelled error
            raise
        if proc.returncode == 0:
            self._logger.info("command succeeded: %s", log_formatted_command)
            return stdout
        else:
            self._logger.warning("command failed: %s", log_formatted_command)
            stderr_str = stderr.decode().lower()
            # Parse some expected errors into specific exceptions
            if "context canceled" in stderr_str:
                error_cls = CommandCancelledError
            # Any error referencing etcd is a connection error
            # This must be before other rules, as it sometimes occurs alonside a not found error
            elif "etcdserver" in stderr_str:
                error_cls = HelmConnectionError
            elif "release: not found" in stderr_str:
                error_cls = ReleaseNotFoundError
            elif "failed to render chart" in stderr_str or "execution error" in stderr_str:
                error_cls = FailedToRenderChartError
            elif "rendered manifests contain a resource that already exists" in stderr_str:
                error_cls = ResourceAlreadyExistsError
            elif "is invalid" in stderr_str:
                error_cls = InvalidResourceError
            elif CHART_NOT_FOUND.search(stderr_str) is not None:
                error_cls = ChartNotFoundError
            elif CONNECTION_ERROR.search(stderr_str) is not None:
                error_cls = HelmConnectionError
            else:
                error_cls = Error
            raise error_cls(proc.returncode, stdout, stderr)

    async def install_or_upgrade(
        self,
        release_name: str,
        chart_ref: pathlib.Path | str,
        values: dict[str, Any] | None = None,
        *,
        atomic: bool = False,
        cleanup_on_fail: bool = False,
        create_namespace: bool = True,
        description: str | None = None,
        devel: bool = False,
        dry_run: bool = False,
        force: bool = False,
        namespace: str | None = None,
        no_hooks: bool = False,
        repo: str | None = None,
        reset_values: bool = False,
        reuse_values: bool = False,
        skip_crds: bool = False,
        timeout: int | str | None = None,
        version: str | None = None,
        wait: bool = False,
        disable_openapi_validation: bool = False,
        skip_schema_validation: bool = False,
    ) -> dict[str, Any]:
        """Installs or upgrades the specified release using the given chart and values."""
        command = [
            "upgrade",
            release_name,
            chart_ref,
            "--history-max",
            self._history_max_revisions,
            "--install",
            "--output",
            "json",
            # Use the default timeout unless an override is specified
            "--timeout",
            timeout if timeout is not None else self._default_timeout,
            # We send the values in on stdin
            "--values",
            "-",
        ]
        if atomic:
            command.append("--atomic")
        if cleanup_on_fail:
            command.append("--cleanup-on-fail")
        if create_namespace:
            command.append("--create-namespace")
        if description:
            command.extend(["--description", description])
        if devel:
            command.append("--devel")
        if dry_run:
            command.append("--dry-run")
        if force:
            command.append("--force")
        if self._insecure_skip_tls_verify:
            command.append("--insecure-skip-tls-verify")
        if namespace:
            command.extend(["--namespace", namespace])
        if no_hooks:
            command.append("--no-hooks")
        if repo:
            command.extend(["--repo", repo])
        if reset_values:
            command.append("--reset-values")
        if reuse_values:
            command.append("--reuse-values")
        if skip_crds:
            command.append("--skip-crds")
        if version:
            command.extend(["--version", version])
        if wait:
            command.extend(["--wait", "--wait-for-jobs"])
        if disable_openapi_validation:
            command.extend(["--disable-openapi-validation"])
        if skip_schema_validation:
            command.extend(["--skip-schema-validation"])
        return json.loads(await self.run(command, json.dumps(values or {}).encode()))

    async def rollback(
        self,
        release_name: str,
        revision: int | None,
        *,
        cleanup_on_fail: bool = False,
        dry_run: bool = False,
        force: bool = False,
        namespace: str | None = None,
        no_hooks: bool = False,
        recreate_pods: bool = False,
        timeout: int | str | None = None,
        wait: bool = False,
    ):
        """Rollback the specified release to the specified revision."""
        command = [
            "rollback",
            release_name,
        ]
        if revision is not None:
            command.append(revision)  # type: ignore
        command.extend(
            [
                "--history-max",
                self._history_max_revisions,
                # Use the default timeout unless an override is specified
                "--timeout",
                timeout if timeout is not None else self._default_timeout,
            ],  # type: ignore
        )
        if cleanup_on_fail:
            command.append("--cleanup-on-fail")
        if dry_run:
            command.append("--dry-run")
        if force:
            command.append("--force")
        if namespace:
            command.extend(["--namespace", namespace])
        if no_hooks:
            command.append("--no-hooks")
        if recreate_pods:
            command.append("--recreate-pods")
        if wait:
            command.extend(["--wait", "--wait-for-jobs"])
        await self.run(command)

    async def show_chart(
        self,
        chart_ref: pathlib.Path | str,
        *,
        devel: bool = False,
        repo: str | None = None,
        version: str | None = None,
    ) -> dict[str, Any]:
        """Returns the contents of Chart.yaml for the specified chart."""
        command = ["show", "chart", chart_ref]
        if devel:
            command.append("--devel")
        if self._insecure_skip_tls_verify:
            command.append("--insecure-skip-tls-verify")
        if repo:
            command.extend(["--repo", repo])
        if version:
            command.extend(["--version", version])
        return yaml.load(await self.run(command), Loader=SafeLoader)

    async def status(  # noqa: ANN201
        self,
        release_name: str,
        *,
        namespace: str | None = None,
        revision: int | None = None,
    ) -> dict[str, Any]:
        """Get the status of the specified release."""
        command = ["status", release_name, "--output", "json"]
        if namespace:
            command.extend(["--namespace", namespace])
        if revision:
            command.extend(["--revision", revision])  # type: ignore
        return json.loads(await self.run(command))

    async def uninstall(
        self,
        release_name: str,
        *,
        dry_run: bool = False,
        keep_history: bool = False,
        namespace: str | None = None,
        no_hooks: bool = False,
        timeout: int | str | None = None,
        wait: bool = False,
    ):
        """Uninstall the specified release."""
        command = [
            "uninstall",
            release_name,
            # Use the default timeout unless an override is specified
            "--timeout",
            timeout if timeout is not None else self._default_timeout,
        ]
        if dry_run:
            command.append("--dry-run")
        if keep_history:
            command.append("--keep-history")
        if namespace:
            command.extend(["--namespace", namespace])
        if no_hooks:
            command.append("--no-hooks")
        if wait:
            command.extend(["--wait"])
        await self.run(command)



from __future__ import annotations

import datetime
import enum
import pathlib
import typing as t
from typing import Annotated, Any, Literal, TypeVar

import yaml
from pydantic import AnyUrl as PydanticAnyUrl
from pydantic import (
    BaseModel,
    DirectoryPath,
    Field,
    FilePath,
    PrivateAttr,
    StringConstraints,
    TypeAdapter,
    field_validator,
)
from pydantic import HttpUrl as PydanticHttpUrl
from pydantic.functional_validators import AfterValidator

from .command import Command, SafeLoader


class ModelWithCommand(BaseModel):
    """Base class for a model that has a Helm command object."""

    # The command object that is used to invoke Helm
    _command: Command = PrivateAttr()

    def __init__(self, _command: Command, **kwargs):
        super().__init__(**kwargs)
        self._command = _command


#: Type for a non-empty string
NonEmptyString = Annotated[
    str,
    StringConstraints(
        min_length=1,
    ),
]

#: Type for a name (chart or release)
Name = Annotated[
    str,
    StringConstraints(
        pattern=r"^[a-z0-9-]+$",
    ),
]


#: Type for a SemVer version
SemVerVersion = Annotated[
    str,
    StringConstraints(
        pattern=r"^v?\d+\.\d+\.\d+(-[a-zA-Z0-9\.\-]+)?(\+[a-zA-Z0-9\.\-]+)?$",
    ),
]


#: Type variables for forward references to the chart and release types
ChartType = TypeVar("ChartType", bound="Chart")
ReleaseType = TypeVar("ReleaseType", bound="Release")


#: Type annotation for validating a string using a Pydantic type
def validate_str_as(validate_type: type):  # noqa
    adapter = TypeAdapter(validate_type)
    return lambda v: str(adapter.validate_python(v))


#: Annotated string types for URLs
AnyUrl = Annotated[str, AfterValidator(validate_str_as(PydanticAnyUrl))]
HttpUrl = Annotated[str, AfterValidator(validate_str_as(PydanticHttpUrl))]
OCIPath = Annotated[str, Field(pattern=r"oci:\/\/*")]


class ChartDependency(BaseModel):
    """Model for a chart dependency."""

    name: Name = Field(
        ...,
        description="The name of the chart.",
    )
    version: NonEmptyString = Field(
        ...,
        description="The version of the chart. Can be a SemVer range.",
    )
    repository: str = Field(
        "",
        description="The repository URL or alias.",
    )
    condition: NonEmptyString | None = Field(
        None,
        description="A yaml path that resolves to a boolean, used for enabling/disabling the chart.",
    )
    tags: list[NonEmptyString] = Field(
        default_factory=list,
        description="Tags can be used to group charts for enabling/disabling together.",
    )
    import_values: list[dict[str, str] | str] = Field(
        default_factory=list,
        alias="import-values",
        description=(
            "Mapping of source values to parent key to be imported. "
            "Each item can be a string or pair of child/parent sublist items."
        ),
    )
    alias: NonEmptyString | None = Field(
        None,
        description="Alias to be used for the chart.",
    )


class ChartMaintainer(BaseModel):
    """Model for the maintainer of a chart."""

    name: NonEmptyString = Field(
        ...,
        description="The maintainer's name.",
    )
    email: NonEmptyString | None = Field(
        None,
        description="The maintainer's email.",
    )
    url: AnyUrl | None = Field(
        None,
        description="A URL for the maintainer.",
    )


class ChartMetadata(BaseModel):
    """Model for chart metadata, from Chart.yaml."""

    api_version: Literal["v1", "v2"] = Field(
        ...,
        alias="apiVersion",
        description="The chart API version.",
    )
    name: Name = Field(
        ...,
        description="The name of the chart.",
    )
    version: SemVerVersion = Field(
        ...,
        description="The version of the chart.",
    )
    kube_version: NonEmptyString | None = Field(
        None,
        alias="kubeVersion",
        description="A SemVer range of compatible Kubernetes versions for the chart.",
    )
    description: NonEmptyString | None = Field(
        None,
        description="A single-sentence description of the chart.",
    )
    type: Literal["application", "library"] = Field(
        "application",
        description="The type of the chart.",
    )
    keywords: list[NonEmptyString] = Field(
        default_factory=list,
        description="List of keywords for the chart.",
    )
    home: HttpUrl | None = Field(
        None,
        description="The URL of th home page for the chart.",
    )
    sources: list[AnyUrl] = Field(
        default_factory=list,
        description="List of URLs to source code for this chart.",
    )
    dependencies: list[ChartDependency] = Field(
        default_factory=list,
        description="List of the chart dependencies.",
    )
    maintainers: list[ChartMaintainer] = Field(
        default_factory=list,
        description="List of maintainers for the chart.",
    )
    icon: HttpUrl | None = Field(
        None,
        description="URL to an SVG or PNG image to be used as an icon.",
    )
    app_version: NonEmptyString | None = Field(
        None,
        alias="appVersion",
        description=("The version of the app that this chart deploys. SemVer is not required."),
    )
    deprecated: bool = Field(
        False,
        description="Whether this chart is deprecated.",
    )
    annotations: dict[str, str] = Field(
        default_factory=dict,
        description="Annotations for the chart.",
    )


class Chart(ModelWithCommand):
    """Model for a reference to a chart."""

    ref: DirectoryPath | FilePath | HttpUrl | Name | OCIPath = Field(
        ...,
        description=(
            "The chart reference. "
            "Can be a chart directory or a packaged chart archive on the local "
            "filesystem, the URL of a packaged chart or the name of a chart, or a oci registry. "
            "When a name is given, repo must also be given and version may optionally "
            "be given."
        ),
    )
    repo: HttpUrl | None = Field(default=None, description="The repository URL.")
    metadata: ChartMetadata = Field(..., description="The metadata for the chart.")

    # Private attributes used to cache attributes
    _readme: str | None = PrivateAttr(None)
    _crds: list[dict[str, Any]] | None = PrivateAttr(None)
    _values: dict[str, Any] | None = PrivateAttr(None)

    @field_validator("ref")
    def ref_is_abspath(cls, v):  # noqa
        """If the ref is a path on the filesystem, make sure it is absolute."""
        if isinstance(v, pathlib.Path):
            return v.resolve()
        else:
            return v

    async def _run_command(self, command_method):  # noqa
        """Runs the specified command for this chart."""
        method = getattr(self._command, command_method)
        # We only need the kwargs if the ref is not a direct reference
        if isinstance(self.ref, pathlib.Path | HttpUrl):  # type: ignore
            return await method(self.ref)
        else:
            return await method(self.ref, repo=self.repo, version=self.metadata.version)


class Release(ModelWithCommand):
    """Model for a Helm release."""

    name: Name = Field(
        ...,
        description="The name of the release.",
    )
    namespace: Name = Field(
        ...,
        description="The namespace of the release.",
    )

    async def current_revision(self) -> ReleaseRevision:
        """Returns the current revision for the release."""
        return ReleaseRevision._from_status(
            await self._command.status(
                self.name,
                namespace=self.namespace,
            ),
            self._command,
        )

    async def revision(self, revision: int) -> ReleaseRevision:
        """Returns the specified revision for the release."""
        return ReleaseRevision._from_status(
            await self._command.status(
                self.name,
                namespace=self.namespace,
                revision=revision,
            ),
            self._command,
        )

    async def rollback(
        self,
        revision: int | None = None,
        *,
        cleanup_on_fail: bool = False,
        dry_run: bool = False,
        force: bool = False,
        no_hooks: bool = False,
        recreate_pods: bool = False,
        timeout: int | str | None = None,
        wait: bool = False,
    ) -> ReleaseRevision:
        """Rollback this release to the specified version and return the resulting revision.

        If no revision is specified, it will rollback to the previous release.
        """
        await self._command.rollback(
            self.name,
            revision,
            cleanup_on_fail=cleanup_on_fail,
            dry_run=dry_run,
            force=force,
            namespace=self.namespace,
            no_hooks=no_hooks,
            recreate_pods=recreate_pods,
            timeout=timeout,
            wait=wait,
        )
        return await self.current_revision()

    async def uninstall(
        self,
        *,
        dry_run: bool = False,
        keep_history: bool = False,
        no_hooks: bool = False,
        timeout: int | str | None = None,
        wait: bool = False,
    ):
        """Uninstalls this release."""
        await self._command.uninstall(
            self.name,
            dry_run=dry_run,
            keep_history=keep_history,
            namespace=self.namespace,
            no_hooks=no_hooks,
            timeout=timeout,
            wait=wait,
        )


class ReleaseRevisionStatus(str, enum.Enum):
    """Enumeration of possible release statuses."""

    #: Indicates that the revision is in an uncertain state
    UNKNOWN = "unknown"
    #: Indicates that the revision has been pushed to Kubernetes
    DEPLOYED = "deployed"
    #: Indicates that the revision has been uninstalled from Kubernetes
    UNINSTALLED = "uninstalled"
    #: Indicates that the revision is outdated and a newer one exists
    SUPERSEDED = "superseded"
    #: Indicates that the revision was not successfully deployed
    FAILED = "failed"
    #: Indicates that an uninstall operation is underway for this revision
    UNINSTALLING = "uninstalling"
    #: Indicates that an install operation is underway for this revision
    PENDING_INSTALL = "pending-install"
    #: Indicates that an upgrade operation is underway for this revision
    PENDING_UPGRADE = "pending-upgrade"
    #: Indicates that a rollback operation is underway for this revision
    PENDING_ROLLBACK = "pending-rollback"


class HookEvent(str, enum.Enum):
    """Enumeration of possible hook events."""

    PRE_INSTALL = "pre-install"
    POST_INSTALL = "post-install"
    PRE_DELETE = "pre-delete"
    POST_DELETE = "post-delete"
    PRE_UPGRADE = "pre-upgrade"
    POST_UPGRADE = "post-upgrade"
    PRE_ROLLBACK = "pre-rollback"
    POST_ROLLBACK = "post-rollback"
    TEST = "test"


class HookDeletePolicy(str, enum.Enum):
    """Enumeration of possible delete policies for a hook."""

    HOOK_SUCCEEDED = "hook-succeeded"
    HOOK_FAILED = "hook-failed"
    HOOK_BEFORE_HOOK_CREATION = "before-hook-creation"


class HookPhase(str, enum.Enum):
    """Enumeration of possible phases for a hook."""

    #: Indicates that a hook is in an unknown state
    UNKNOWN = "Unknown"
    #: Indicates that a hook is currently executing
    RUNNING = "Running"
    #: Indicates that hook execution succeeded
    SUCCEEDED = "Succeeded"
    #: Indicates that hook execution failed
    FAILED = "Failed"


class Hook(BaseModel):
    """Model for a hook."""

    name: NonEmptyString = Field(
        ...,
        description="The name of the hook.",
    )
    phase: HookPhase = Field(
        HookPhase.UNKNOWN,
        description="The phase of the hook.",
    )
    kind: NonEmptyString = Field(
        ...,
        description="The kind of the hook.",
    )
    path: NonEmptyString = Field(
        ...,
        description="The chart-relative path to the template that produced the hook.",
    )
    resource: dict[str, t.Any] = Field(
        ...,
        description="The resource for the hook.",
    )
    events: list[HookEvent] = Field(
        default_factory=list,
        description="The events that the hook fires on.",
    )
    delete_policies: list[HookDeletePolicy] = Field(
        default_factory=list,
        description="The delete policies for the hook.",
    )


class ReleaseRevision(ModelWithCommand):
    """Model for a revision of a release."""

    release: Release = Field(
        ...,
        description="The parent release of this revision.",
    )
    revision: int = Field(
        ...,
        description="The revision number of this revision.",
    )
    status: ReleaseRevisionStatus = Field(
        ...,
        description="The status of the revision.",
    )
    updated: datetime.datetime = Field(
        ...,
        description="The time at which this revision was updated.",
    )
    description: NonEmptyString | None = Field(
        None,
        description="'Log entry' for this revision.",
    )
    notes: NonEmptyString | None = Field(
        None,
        description="The rendered notes for this revision, if available.",
    )

    # Optional fields if they are known at creation time
    chart_metadata_: ChartMetadata | None = Field(None, alias="chart_metadata")
    hooks_: list[dict[str, t.Any]] | None = Field(None, alias="hooks")
    resources_: list[dict[str, t.Any]] | None = Field(None, alias="resources")
    values_: dict[str, t.Any] | None = Field(None, alias="values")

    def _set_from_status(self, status: dict[str, Any]):
        # Statuses from install/upgrade have chart metadata embedded
        if "chart" in status:
            self.chart_metadata_ = ChartMetadata(**status["chart"]["metadata"])
        self.hooks_ = [  # type: ignore
            Hook(
                name=hook["name"],
                phase=hook["last_run"].get("phase") or "Unknown",  # type: ignore
                kind=hook["kind"],
                path=hook["path"],
                resource=yaml.load(hook["manifest"], Loader=SafeLoader),
                events=hook["events"],
                delete_policies=hook.get("delete_policies", []),
            )
            for hook in status.get("hooks", [])
        ]
        self.resources_ = list(yaml.load_all(status["manifest"], Loader=SafeLoader))

    @classmethod
    def _from_status(cls, status: dict[str, Any], command: Command) -> ReleaseRevision:
        """Internal constructor to create a release revision from a status result."""
        revision = ReleaseRevision(
            command,  # type: ignore
            release=Release(
                command,  # type: ignore
                name=status["name"],
                namespace=status["namespace"],
            ),
            revision=status["version"],
            status=status["info"]["status"],
            updated=status["info"]["last_deployed"],
            description=status["info"].get("description"),
            notes=status["info"].get("notes"),
        )
        revision._set_from_status(status)
        return revision



import pytest

from dagster_uc.config import load_config


@pytest.mark.parametrize(
    ("environment", "expected"),
    [
        (
            "dev",
            {
                "repository_root": ".",
                "environment": "dev",
                "code_path": "example/repo.py",
                "dagster_version": "1.11.16",
                "cicd": False,
                "dagster_gui_url": "http://dagster.dev",
                "verbose": False,
                "use_project_name": True,
                "use_latest_chart_version": True,
                "docker_config": {
                    "container_registry": "dagster-uc.dev.acr.io",
                    "dockerfile": "docker/dev.Dockerfile",
                    "docker_root": ".",
                    "docker_env_vars": ["FOO", "FOO='string'"],
                    "image_prefix": "example",
                    "use_az_login": True,
                    "container_registry_chart_path": "helm/dagster/dagster-user-deployments",
                    "build_format": "OCI",
                },
                "helm_config": {
                    "disable_openapi_validation": False,
                    "skip_schema_validation": True,
                    "create_new_namespace": True,
                },
                "kubernetes_config": {
                    "context": "aks-dev",
                    "namespace": "dagster",
                    "node": "cpunode",
                    "image_pull_secrets": [],
                    "user_code_deployment_env_secrets": [{"name": "dagster-storage-secret"}],
                    "user_code_deployment_env": [
                        {"name": "ON_K8S", "value": "1"},
                        {"name": "ENVIRONMENT", "value": "dev"},
                    ],
                    "limits": {"cpu": "4000m", "memory": "2000Mi"},
                    "requests": {"cpu": "150m", "memory": "750Mi"},
                },
                "dagster_chart_config": {
                    "deployments_configmap_name": "dagster-user-deployments-values-yaml",
                    "workspace_yaml_configmap_name": "dagster-workspace-yaml",
                    "deployment_semaphore_name": "dagster-uc-semaphore",
                    "release_name": "dagster-user-code",
                },
            },
        ),
        (
            "acc",
            {
                "repository_root": ".",
                "environment": "acc",
                "code_path": "example/repo.py",
                "dagster_version": "1.11.16",
                "cicd": False,
                "dagster_gui_url": "http://dagster.acc",
                "verbose": False,
                "use_project_name": True,
                "use_latest_chart_version": True,
                "docker_config": {
                    "container_registry": "dagster-uc.acc.acr.io",
                    "dockerfile": "docker/acc.Dockerfile",
                    "docker_root": ".",
                    "docker_env_vars": ["FOO", "FOO='string'"],
                    "image_prefix": "example",
                    "use_az_login": True,
                    "container_registry_chart_path": "helm/dagster/dagster-user-deployments",
                    "build_format": "OCI",
                },
                "helm_config": {
                    "disable_openapi_validation": False,
                    "skip_schema_validation": True,
                    "create_new_namespace": True,
                },
                "kubernetes_config": {
                    "context": "aks-acc",
                    "namespace": "dagster",
                    "node": "cpunode",
                    "image_pull_secrets": [],
                    "user_code_deployment_env_secrets": [{"name": "dagster-storage-secret"}],
                    "user_code_deployment_env": [
                        {"name": "ON_K8S", "value": "1"},
                        {"name": "ENVIRONMENT", "value": "acc"},
                    ],
                    "limits": {"cpu": "4000m", "memory": "2000Mi"},
                    "requests": {"cpu": "150m", "memory": "750Mi"},
                },
                "dagster_chart_config": {
                    "deployments_configmap_name": "dagster-user-deployments-values-yaml",
                    "workspace_yaml_configmap_name": "dagster-workspace-yaml",
                    "deployment_semaphore_name": "dagster-uc-semaphore",
                    "release_name": "dagster-user-code",
                },
            },
        ),
    ],
)
def test_load_config(environment: str, expected: dict):  # noqa
    dev_config = load_config(environment, path="tests/config/.config_user_code_deployments.yaml")

    assert dev_config.model_dump() == expected



from __future__ import annotations

import os
from copy import deepcopy
from pathlib import Path
from string import Template
from typing import Any, Literal

from pydantic import BaseModel, Field
from pydantic_settings import (
    BaseSettings,
    EnvSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)

from dagster_uc.log import logger

OVERRIDE_CONFIG_FILE_PATH: str | None = None
DEFAULT_CONFIG_FILE_PATH = ".config_user_code_deployments.yaml"


def get_config_file_path() -> str:
    """Gets config file path in this order, manual input from cli -> env var -> default"""
    if OVERRIDE_CONFIG_FILE_PATH is None:
        return os.environ.get("CONFIG_FILE_PATH", DEFAULT_CONFIG_FILE_PATH)
    else:
        return OVERRIDE_CONFIG_FILE_PATH


class KubernetesResource(BaseModel):
    """Kubernetes resources"""

    cpu: str | int
    memory: str | int


class KubernetesSecret(BaseModel):
    """Kubernetes secret"""

    name: str


class KubernetesEnvVar(BaseModel):
    """Kubernetes env var"""

    name: str
    value: str


def deep_merge(a: dict[str, Any], b: dict[str, Any]) -> dict[str, Any]:
    """Recursively merge dict `b` into `a`."""
    result = deepcopy(a)
    for key, value in b.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = deep_merge(result[key], value)
        else:
            result[key] = value
    return result


class EnvParsedYamlConfigSettingsSource(YamlConfigSettingsSource):
    """Yaml Settings Source that parses envVar references before loading. Also selects the env subset based one the environment.
    And merges the defaults in as well.

    ENV VARS need to be configured like this: ${ENV_VAR}

    An example of a valid config.yaml would be:
        ```yaml
        defaults:
            var: "foo"

        dev:
            var: "${env_var}"
        acc:
            var2: "bar"
        ```
    """

    def _read_file(self, file_path: Path) -> dict[str, Any]:
        import json

        data = super()._read_file(file_path)

        environment = os.environ.get("ENVIRONMENT", "dev")

        # get based on environment
        environment_data = data.get(environment)

        if environment_data is None:
            logger.warning("Didn't find environment %s in config yaml", environment)
            environment_data = {}

        defaults_data = data.get("defaults", {})
        combined_data = deep_merge(defaults_data, environment_data)

        data = json.dumps(combined_data)
        data = Template(data).substitute(os.environ)
        return json.loads(data)


class _BaseSettingsWithYamlAndEnv(BaseSettings):
    model_config = SettingsConfigDict(
        extra="ignore",
    )

    @classmethod
    def settings_customise_sources(  # noqa: D102
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        env_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        dotenv_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        file_secret_settings: PydanticBaseSettingsSource,  # noqa: ARG003
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            EnvSettingsSource(settings_cls),
            EnvParsedYamlConfigSettingsSource(settings_cls, yaml_file=get_config_file_path()),
        )


class DockerConfiguration(BaseModel):
    """Configuration specific for Docker or Container registry related."""

    container_registry: str
    dockerfile: str
    docker_root: str = Field(default="")
    docker_env_vars: list[str] = Field(default=[])
    image_prefix: str = Field(default="")
    use_az_login: bool = Field(default=True)
    container_registry_chart_path: str | None = Field(default=None)
    build_format: Literal["OCI", "docker"] = "OCI"


class HelmConfiguration(BaseModel):
    """Helm specific configuration"""

    disable_openapi_validation: bool = Field(default=False)
    skip_schema_validation: bool = Field(default=False)
    create_new_namespace: bool = Field(default=True)


class KubernetesConfiguration(BaseModel):
    """Kubernetes specific configuration."""

    context: str
    namespace: str
    node: str
    image_pull_secrets: list[KubernetesSecret] = Field(default=[])
    user_code_deployment_env_secrets: list[KubernetesSecret] = Field(default=[])
    user_code_deployment_env: list[KubernetesEnvVar] = Field(default=[])
    limits: KubernetesResource | None = Field(default=None)
    requests: KubernetesResource | None = Field(default=None)


class DagsterUserCodeChartConfiguration(BaseModel):
    """User code chart configuration"""

    deployments_configmap_name: str = "dagster-user-deployments-values-yaml"
    workspace_yaml_configmap_name: str = "dagster-workspace-yaml"
    deployment_semaphore_name: str = "dagster-uc-semaphore"
    release_name: str = "dagster-user-code"


class DagsterUserCodeConfiguration(_BaseSettingsWithYamlAndEnv):
    """Dagster User Code configuration"""

    # Environment/code config
    repository_root: str = Field(default="")
    environment: str
    code_path: str
    dagster_version: str
    cicd: bool = Field(default=False)
    dagster_gui_url: str | None = Field(default=None)
    verbose: bool = Field(default=False)
    use_project_name: bool = Field(default=True)
    use_latest_chart_version: bool = False

    docker_config: DockerConfiguration = Field(default_factory=DockerConfiguration, alias="docker")  # type: ignore
    helm_config: HelmConfiguration = Field(default_factory=HelmConfiguration, alias="helm")
    kubernetes_config: KubernetesConfiguration = Field(
        default_factory=KubernetesConfiguration,  # type: ignore
        alias="kubernetes",
    )
    dagster_chart_config: DagsterUserCodeChartConfiguration = Field(
        default_factory=DagsterUserCodeChartConfiguration,
        alias="chart",
    )


def load_config(environment: str, path: str | None = None) -> DagsterUserCodeConfiguration:
    """Loads the configuration file from the local dir or the user's home dir."""
    os.environ["ENVIRONMENT"] = environment
    global OVERRIDE_CONFIG_FILE_PATH
    if path is not None:
        OVERRIDE_CONFIG_FILE_PATH = path

    return DagsterUserCodeConfiguration()  # type: ignore



from __future__ import annotations


class Error(Exception):
    """Raised when an error occurs with a Helm command."""

    def __init__(self, returncode: int | None, stdout: bytes, stderr: bytes):
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        super().__init__(stderr.decode())


class HelmConnectionError(Error):
    """Raised when there is a problem connecting to the Kubernetes API."""


class ChartNotFoundError(Error):
    """Raised when a chart is not found."""


class FailedToRenderChartError(Error):
    """Raised when a chart fails to render."""


class ReleaseNotFoundError(Error):
    """Raised when a release is not found."""


class ResourceAlreadyExistsError(Error):
    """Raised when Helm attempts to create a resource that already exists."""


class InvalidResourceError(Error):
    """Raised when Helm attempts to create or update a resource in a way that is not valid."""


class CommandCancelledError(Error):
    """Raised when a Helm command is cancelled."""



from __future__ import annotations

import contextlib
import json
import logging
import os
import pprint
import time
from typing import Annotated, cast

import kr8s
import typer
from kr8s.objects import (
    Pod,
)

from dagster_uc.config import (
    DagsterUserCodeChartConfiguration,
    DagsterUserCodeConfiguration,
    DockerConfiguration,
    KubernetesConfiguration,
    load_config,
)
from dagster_uc.log import logger
from dagster_uc.uc_handler import DagsterUserCodeHandler
from dagster_uc.utils import (
    BuildTool,
    build_and_push,
    gen_tag,
    is_command_available,
)

app = typer.Typer(invoke_without_command=True)
deployment_app = typer.Typer(
    name="deployment",
    help="Contains various subcommands for managing user code deployments",
    no_args_is_help=True,
)
app.add_typer(deployment_app)
deployment_delete_app = typer.Typer(
    name="delete",
    help="Delete one or more user code deployments",
)
deployment_app.add_typer(deployment_delete_app)
deployment_check_app = typer.Typer(
    name="check",
    help="Check status of the user code deployment",
)
deployment_app.add_typer(deployment_check_app)

handler: DagsterUserCodeHandler
config: DagsterUserCodeConfiguration


@app.command("show-config", help="Outputs the configuration that is currently in use")
def show_config():
    """Pretty print the config object"""
    pprint.pprint(config, indent=4)


@app.command(
    name="init-config",
    help="Create a config for dagster deployment",
    no_args_is_help=True,
)
def init_config(
    file: Annotated[
        str,
        typer.Option(
            "--file",
            "-f",
            help="File and path where to save the config example.",
        ),
    ],
):
    """Initiates a config template"""

    def optional_prompt(text: str) -> str | None:
        var = typer.prompt(text, default="")
        if var:
            return var
        else:
            return None

    initialized_config = DagsterUserCodeConfiguration.model_construct(
        environment=typer.prompt("""What environment is this config for [dev, acc, prod etc.]"""),
        code_path=typer.prompt("Path of the Definitions python file inside docker image"),
        repository_root=typer.prompt("Path of project scope", default="."),
        dagster_version=typer.prompt(
            "Version of dagster in project (should mirror dagster depoyment!)",
        ),
        cicd=typer.confirm(
            "Whether it's executed in CICD. If set to True, then the deployment_name is created from the env",
        ),
        dagster_gui_url=optional_prompt("URL of dagster UI"),
        use_latest_chart_version=typer.confirm(
            "Whether to use latest chart version, if disabled will always use the config dagster version",
        ),
        use_project_name=typer.confirm(
            "Whether to use the pyproject.toml project-name as deployment name prefix.",
        ),
        docker=DockerConfiguration(
            docker_root=typer.prompt("Path of docker scope", default="."),
            dockerfile=typer.prompt("Path of dockerfile", default="./Dockerfile"),
            image_prefix=typer.prompt("Prefix for container image to use"),
            container_registry=typer.prompt("Container registry address"),
            container_registry_chart_path=optional_prompt(
                "Relative helm chart path of the previously provided container registry",
            ),
            use_az_login=typer.confirm(
                "Whether to use az cli to login to container registry with podman, and helm if oci charts are used.",
            ),
        ),
        kubernetes=KubernetesConfiguration(
            context=typer.prompt("Kubernetes context of the cluster to use for api calls"),
            namespace=typer.prompt("Namespace of kubernetes deployment"),
            node=typer.prompt("Kubernetes node for user-code pod"),
            requests=json.loads(
                typer.prompt(
                    "Request for the user pod in k8s in json formatted string",
                    default=json.dumps({"cpu": "1", "memory": "1Gi"}),
                ),
            ),
            limits=json.loads(
                typer.prompt(
                    "Limits for the user pod in k8s in json formatted string",
                    default=json.dumps({"cpu": "2", "memory": "2Gi"}),
                ),
            ),
        ),
        chart=DagsterUserCodeChartConfiguration(
            deployments_configmap_name=typer.prompt(
                "Configmap name to use for user_code_deployments",
                default="dagster-user-deployments-values-yaml",
            ),
            workspace_yaml_configmap_name=typer.prompt(
                "Configmap name of the dagster_workspace_yaml",
                default="dagster-workspace-yaml",
            ),
        ),
    )

    with open(file, "w") as fp:
        import yaml

        config_dict = initialized_config.model_dump(by_alias=True)
        complete_dict = {config_dict["environment"]: config_dict}
        yaml.dump(complete_dict, fp, default_flow_style=False, sort_keys=False)
    typer.echo(f"Template configuration file generated as '{file}'.")


@app.callback(invoke_without_command=True, no_args_is_help=True)
def default(
    ctx: typer.Context,
    environment: str = typer.Option("dev", "--environment", "-e", help="The environment"),
    config_file_path: str = typer.Option(
        None,
        "--config-file",
        "-c",
        help="Path to the config file.",
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Output DEBUG logging"),
) -> None:
    """This function executes before any other nested cli command is called and loads the configuration object."""
    global logger
    global config
    global handler

    if ctx.invoked_subcommand == "init-config":
        pass
    else:
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)
        config = load_config(environment, config_file_path)

        if verbose:
            config.verbose = True
        logger.debug(f"Switching kubernetes context to {config.environment}...")
        kr8s_api = kr8s.api(
            context=f"{config.kubernetes_config.context}",
            namespace=config.kubernetes_config.namespace,
        )

        handler = DagsterUserCodeHandler(config, kr8s_api)
        handler.maybe_create_user_deployments_configmap()
        logger.debug(f"Done: Switched kubernetes context to {config.environment}")


def build_push_container(
    deployment_name: str,
    branch_name: str,
    image_prefix: str | None,
    config: DagsterUserCodeConfiguration,
    use_sudo: bool,
    tag: str,
) -> None:
    """Builds a docker image for a user-code deployment of the current branch and uploads it to the image registry"""
    handler.update_dagster_workspace_yaml()
    build_and_push(
        config.repository_root,
        config.docker_config.container_registry,
        image_name=deployment_name
        if not image_prefix
        else os.path.join(image_prefix, deployment_name),
        dockerfile=config.docker_config.dockerfile,
        use_sudo=use_sudo,
        tag=tag,
        branch_name=branch_name,
        use_az_login=config.docker_config.use_az_login,
        build_envs=config.docker_config.docker_env_vars,
        build_format=config.docker_config.build_format,
    )


@deployment_app.command(
    name="list",
    help="List user code deployments that are currently active",
)
def deployment_list():
    """Outputs a list of currently active deployments"""
    typer.echo(
        "\033[1mActive user code deployments\033[0m\n"
        + "\n".join(
            ["* " + d["name"] for d in handler.list_deployments()],
        ),
    )


@deployment_app.command(
    name="revive",
    help="Redeploy an old user code deployment using the latest existing image from the container registry",
)
def deployment_revive(
    name: Annotated[
        str,
        typer.Option("--name", "-n", help="The name of the deployment to revive."),
    ],
    tag: Annotated[str, typer.Option("--tag", "-t", help="The tag of the deployment to revive.")],
):
    """Revives an old deployment from the container registry."""
    # In case the UI name separator of the deployment is passed
    name = name.replace(":", "--")

    if not handler._check_deployment_exists(
        name,
    ):
        handler.add_user_deployment_to_configmap(
            handler.gen_new_deployment_yaml(
                name,
                image_prefix=handler.config.docker_config.image_prefix,
                tag=tag,
            ),
        )
        handler.deploy_to_k8s()
        logger.info(f"Deployed {name}")
    else:
        raise Exception(f'Deployment "{name}" already exists')


@deployment_delete_app.callback(invoke_without_command=True)
def deployment_delete(
    delete_all: Annotated[
        bool,
        typer.Option(
            "--all",
            "-a",
            help="If this is provided, all deployments will be deleted including CI/CD provided deployments.",
        ),
    ] = False,
    name: Annotated[
        str | None,
        typer.Option(
            "--name",
            "-n",
            help="The name of the deployment to delete. If none is provided, the deployment corresponding to the currently checked out git branch will be deleted.",
        ),
    ] = None,
    branch: Annotated[
        str | None,
        typer.Option(
            "--branch",
            "-b",
            help="Can be the branch name as-is, this will be converted to a dagster-uc compatible deployment name.",
        ),
    ] = None,
) -> None:
    """Delete a registered user-code deployment from kubernetes."""
    if delete_all:
        handler.remove_all_deployments()
        handler.deploy_to_k8s(reload_dagster=True)
        typer.echo("\033[1mDeleted all deployments\033[0m")
    else:
        if name is not None:
            # In case the UI name separator of the deployment is passed
            name = name.replace(":", "--")
        elif branch is not None:
            name = handler.get_deployment_name(
                use_project_name=config.use_project_name,
                branch=branch,
            ).full_name
        else:
            name = handler.get_deployment_name(
                use_project_name=config.use_project_name,
            ).full_name

        handler.remove_user_deployment_from_configmap(name)
        handler.deploy_to_k8s(reload_dagster=True)
        typer.echo(f"Deleted deployment \033[1m{name}\033[0m")


@deployment_check_app.callback(invoke_without_command=True)
def check_deployment(
    name: Annotated[
        str,
        typer.Option(
            "--name",
            "-n",
            help="The name of the deployment to check. If not provided, checks deployment corresponding to the current branch.",
        ),
    ] = "",
    timeout: Annotated[
        int,
        typer.Option(
            "--timeout",
            "-t",
            help="The timeout duration in seconds to keep following the logs of user code pod.",
        ),
    ] = 60,
) -> None:
    """This function executes before any other nested cli command is called and loads the configuration object."""
    if not name:
        name = handler.get_deployment_name(use_project_name=config.use_project_name).full_name
    else:
        # In case the UI name separator of the deployment is passed
        name = name.replace(":", "--")
    if not handler._check_deployment_exists(name):
        logger.warning(
            f"Deployment with name '{name}' does not seem to exist in environment '{config.environment}'. Attempting to proceed with status check anyways.",
        )
    typer.echo(f"\033[1mStatus for deployment {name}\033[0m")
    for pod in handler.api.get(
        Pod,
        label_selector=f"deployment={name}",
        namespace=config.kubernetes_config.namespace,
    ):
        pod = cast(Pod, pod)
        with contextlib.suppress(Exception):
            for line in pod.logs(pretty=True, follow=True, timeout=timeout):  # type: ignore
                typer.echo(line)
                if "started dagster code server" in line.lower():
                    break


@deployment_app.command(
    name="deploy",
    help="Deploys the currently checked out git branch as a user code deployment",
)
def deployment_deploy(
    force: Annotated[
        bool,
        typer.Option(
            "--force",
            "-f",
            help="If this is provided, a full redeploy will always be done, rather than just rebooting user code pods if they already exist in order to trigger a new image pull",
        ),
    ] = False,
    skip_build: bool = typer.Option(
        False,
        "--skip-build",
        "-b",
        help="Image build and push will be skipped",
    ),
    deployment_name_suffix: str = typer.Option(
        "",
        "--deployment-name-suffix",
        "-s",
        help="Suffix to append to the default deployment name (which is based on the git branch name) - useful for doing parallel deployments from the same branch.",
    ),
    deployment_name: str = typer.Option(
        "",
        "--deployment-name",
        "-n",
        help="Overrides the name of the deployment, including any --deployment-name-suffix value",
    ),
    reset_lock: bool = typer.Option(
        False,
        "--reset-lock",
        "-r",
        help="Reset the deployment semaphore of any ongoing other deployments.",
    ),
    use_sudo: Annotated[
        bool,
        typer.Option(
            "--use-sudo",
            "-u",
            help="If this is provided, buildah or docker will be called with sudo",
        ),
    ] = False,
    ignore_check: Annotated[
        bool,
        typer.Option(
            "--ignore-check",
            help="If this is provided, ignore the check for if podman is installed. This is used for some CI/CD environments that require podman to always be in sudo mode",
        ),
    ] = False,
):
    """Handles deployment to kubernetes cluster."""
    handler._ensure_dagster_version_match()

    count = 0
    while not handler.acquire_semaphore(reset_lock):
        logger.error(
            f"Attempt {count}: Another deployment is in progress. Trying again in 10 seconds. You can force a reset of the deployment lock by using 'dagster-uc deployment deploy --reset-lock'",
        )
        count += 1
        time.sleep(10)

    try:
        logger.debug("Determining build tool...")
        if not ignore_check and not is_command_available(BuildTool.podman.value):
            raise Exception("Podman installation is required to run dagster-uc.")

        logger.debug("Using 'podman' to build image.")
        if deployment_name:
            (deployment_name, branch_name) = (deployment_name, deployment_name)
        else:
            dagster_deployment = handler.get_deployment_name(
                deployment_name_suffix,
                use_project_name=config.use_project_name,
            )
            (deployment_name, branch_name) = (
                dagster_deployment.full_name,
                dagster_deployment.branch_name,
            )

        logger.debug("Determining tag...")
        new_tag = gen_tag(
            deployment_name
            if not config.docker_config.image_prefix
            else os.path.join(config.docker_config.image_prefix, deployment_name),
            config.docker_config.container_registry,
            config.dagster_version,
            config.docker_config.use_az_login,
            use_sudo=use_sudo,
        )

        typer.echo(f"Deploying deployment \033[1m'{deployment_name}:{new_tag}'\033[0m")

        full_redeploy_done = False
        if not skip_build:
            build_push_container(
                deployment_name,
                branch_name=branch_name,
                image_prefix=handler.config.docker_config.image_prefix,
                config=config,
                use_sudo=use_sudo,
                tag=new_tag,
            )

        if not handler._check_deployment_exists(deployment_name):
            logger.info(
                f"Deployment with name '{deployment_name}' does not exist yet in '{config.environment}'. Adding deployment to configmap",
            )
            handler.add_user_deployment_to_configmap(
                handler.gen_new_deployment_yaml(
                    deployment_name,
                    image_prefix=handler.config.docker_config.image_prefix,
                    tag=new_tag,
                ),
            )
            handler.deploy_to_k8s()
        else:
            logger.info(
                f"Deployment with name '{deployment_name}' exists in '{config.environment}'. Updating deployment in configmap",
            )
            # TODO(ion): make this into a single operation (replace)
            handler.remove_user_deployment_from_configmap(deployment_name)
            handler.add_user_deployment_to_configmap(
                handler.gen_new_deployment_yaml(
                    deployment_name,
                    image_prefix=handler.config.docker_config.image_prefix,
                    tag=new_tag,
                ),
            )
            if config.cicd or force:
                handler.deploy_to_k8s()
            elif not handler.check_if_code_pod_exists(label=deployment_name):
                logger.info(
                    "Code deployment present in configmap but pod not found, triggering full deploy...",
                )
                handler.deploy_to_k8s()  # Something went wrong - redeploy yamls and reload webserver
            else:
                logger.info(
                    "Code deployment present in configmap and pod found...",
                )
                handler.deploy_to_k8s(reload_dagster=False)
    finally:
        handler.release_semaphore()
    if config.dagster_gui_url:
        typer.echo(
            f"Your assets: {config.dagster_gui_url.rstrip('/')}/locations/{deployment_name.replace('--', ':')}/assets\033[0m",
        )
    time.sleep(5)
    timeout = 40 if not full_redeploy_done else 240

    while True:
        code_pods = list(
            handler.api.get(
                Pod,
                label_selector=f"deployment={deployment_name}",
                namespace=config.kubernetes_config.namespace,
            ),
        )
        if len(code_pods) == 0:
            time.sleep(2)
            continue
        else:
            break

    with contextlib.suppress(Exception):
        for code_pod in code_pods:
            code_pod.wait("condition=Ready", timeout=timeout)  # type: ignore
    check_deployment(deployment_name)


if __name__ == "__main__":
    app()



from __future__ import annotations

import functools
import pathlib
from typing import Any

from dagster_uc._helm.command import Command
from dagster_uc._helm.models import Chart, ReleaseRevision


def mergeconcat(
    defaults: dict[Any, Any],
    *overrides: dict[Any, Any],
) -> dict[Any, Any]:
    """Deep-merge two or more dictionaries together. Lists are concatenated."""

    def mergeconcat2(defaults, overrides):  # noqa: ANN001, ANN202
        if isinstance(defaults, dict) and isinstance(overrides, dict):
            merged = dict(defaults)
            for key, value in overrides.items():
                if key in defaults:
                    merged[key] = mergeconcat2(defaults[key], value)
                else:
                    merged[key] = value
            return merged
        elif isinstance(defaults, list | tuple) and isinstance(overrides, list | tuple):
            merged = list(defaults)
            merged.extend(overrides)
            return merged
        else:
            return overrides if overrides is not None else defaults

    return functools.reduce(mergeconcat2, overrides, defaults)  # type: ignore


class Client:
    """Entrypoint for interactions with Helm."""

    def __init__(
        self,
        *,
        default_timeout: int | str = "5m",
        executable: str = "helm",
        history_max_revisions: int = 10,
        insecure_skip_tls_verify: bool = False,
        kubeconfig: pathlib.Path | None = None,
        kubecontext: str | None = None,
        unpack_directory: str | None = None,
    ):
        self._command = Command(
            default_timeout=default_timeout,
            executable=executable,
            history_max_revisions=history_max_revisions,
            insecure_skip_tls_verify=insecure_skip_tls_verify,
            kubeconfig=kubeconfig,
            kubecontext=kubecontext,
            unpack_directory=unpack_directory,
        )

    async def get_chart(
        self,
        chart_ref: pathlib.Path | str,
        *,
        devel: bool = False,
        repo: str | None = None,
        version: str | None = None,
    ) -> Chart:
        """Returns the resolved chart for the given ref, repo and version."""
        return Chart(
            self._command,  # type: ignore
            ref=chart_ref,
            repo=repo,
            # Load the metadata for the specified args
            metadata=await self._command.show_chart(
                chart_ref,
                devel=devel,
                repo=repo,
                version=version,
            ),
        )

    async def install_or_upgrade_release(
        self,
        release_name: str,
        chart: Chart,
        *values: dict[str, Any],
        atomic: bool = False,
        cleanup_on_fail: bool = False,
        create_namespace: bool = True,
        description: str | None = None,
        dry_run: bool = False,
        force: bool = False,
        namespace: str | None = None,
        no_hooks: bool = False,
        reset_values: bool = False,
        reuse_values: bool = False,
        skip_crds: bool = False,
        timeout: int | str | None = None,
        wait: bool = False,
        disable_openapi_validation: bool = False,
        skip_schema_validation: bool = False,
    ) -> ReleaseRevision:
        """Install or upgrade the named release using the given chart and values and return
        the new revision.
        """
        return ReleaseRevision._from_status(
            await self._command.install_or_upgrade(
                release_name,
                chart.ref,
                mergeconcat(*values) if values else None,
                atomic=atomic,
                cleanup_on_fail=cleanup_on_fail,
                create_namespace=create_namespace,
                description=description,
                dry_run=dry_run,
                force=force,
                namespace=namespace,
                no_hooks=no_hooks,
                repo=chart.repo,
                reset_values=reset_values,
                reuse_values=reuse_values,
                skip_crds=skip_crds,
                timeout=timeout,
                version=chart.metadata.version,
                wait=wait,
                disable_openapi_validation=disable_openapi_validation,
                skip_schema_validation=skip_schema_validation,
            ),
            self._command,
        )
